/*
  Name:        	TraceVerify.c
  Author:      	Ron Burkey
  Purpose:     	The processing for Do178Builder-utility 
  		that creates the traceability matrix for the
		--trace-verify command-line switch.
  Mod History:	03/14/2008 RSB	Began.
  		03/17/2008 RSB	Adapted from TraceDesign.c.
		03/19/2008 RSB	Fixed a bug in which the section
				count wasn't getting initialized
				to zero after the high-level 
				requirement was closed out, causing
				missing lines in the table.
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  
  The dataset output by this function is a CSV file with one record 
  for each pairing of a high-level or low-level requirement with an
  appropriate section in the SVCP document.  A requirement which has 
  no such association is matched with "(none)".  For example, suppose 
  that high-level requirement H1 is associated with SVCP section 
  1.2.3.4.5, while low-level requirements L1 and L2, are associated
  with SVCP section 1.2.3.4.6.  Then the output file would look 
  something like the following, with the first line containing
  column headers:
  
  	"Requirement","Test Case"
  	"SRD H1","SVCP 1.2.3.4.5"
	"SDD L1","SVCP 1.2.3.4.6"
	"SDD L2","SVCP 1.2.3.4.6"

  You might think that the notations "SRD", "SDD", and "SVCP" above 
  are pointless (e.g., why "SRD H1" rather than just "H1"?), but reason 
  they are there is to allow easy combination of CSV data created from 
  different databases into a single table.  For example, one could 
  combine data from a DO-178B database and a DO-254 database, and
  the face that some items read "SRD H1" while others read "HRD H1"
  would allow the requirements from the different databases to 
  remain distinct.

  The requirements and SVCP sections are identified by pattern 
  matching with user-supplied regular expressions. 
  
  By sucking the CSV file into a spreadsheet program, it can be sorted 
  in order of the high-level requirements or in order of the low-level
  requirements.  The actual output order is simply the order in which
  the requirements are found in the XML.
*/

#include "Do178Builder-utility.h"

#define CsvFilename "Do178Builder-trace-verify.csv"
// The reason the following lines are here is that this file was originally
// adapted from TraceDesign.c, and we want to make sure that we've changed
// the names of all data structures referring to --trace-design, so we
// reassign them to nonsense words that would result in compile-time
// errors.
#define TraceDesignState eeny
#define TraceDesignCloseout meeny
#define FileTraceDesign miny
#define TraceDesign moe


static void
TraceVerifyCloseoutSpecific (ProcessingOptions_t *Options, TraceVerifyState_t *ts)
{
  int k;
  if (ts == NULL)
    return;
  if (!ts->TestCaseCount)
    {
      for (k = 0; k < ts->NumRequirements; k++)
        {
	  fprintf (Options->FileTraceVerify, "\"%s %s\",\"%s %s\",\"%s\"\n",
		   ts->DocName,
		   ts->Requirements[k],
		   ts->DocName,
		   ts->RequirementSection.Number,
		   "(none)");
        }
#if 0	
      if (k == 0)
        {
	  fprintf (Options->FileTraceVerify, "\"%s\",\"%s %s\",\"%s\"\n",
		   "(none)",
		   ts->DocName,
		   ts->RequirementSection.Number,
		   "(none)");
	}
#endif	
    }
  ts->TestCaseCount = 0;
  ts->NumRequirements = 0;
}

// Close out any test-case collection in progress at this (high or low) level.
void
TraceVerifyCloseout (ProcessingOptions_t *Options)
{
  
  // Make sure low-level requirement is closed out.
  TraceVerifyCloseoutSpecific (Options, &Options->TraceVerifyState);
  
  // Close out high-level requirement if need be.
  if (Options->WalkLevel == Options->VerifyParentWalkLevel)
    TraceVerifyCloseoutSpecific (Options, &Options->TraceVerifyStateParent);
        
}

// This function does the processing by being called by WalkNodes
// once for each node in the XML tree.  It has to maintain its own
// state machine in order to tie all of these separate calls together,
// and the state machine comprises fields in Options.  Calling with
// a_node==NULL should initialize (reset) the state machine.
int 
TraceVerify (xmlNode *a_node, ProcessingOptions_t *Options)
{
  int i, j, k, n;
  char c;
  int CanContainHighLevel = 0, CanContainLowLevel = 0;
  int DoesContainHighLevel = 0, DoesContainLowLevel = 0;
  int CanContainVerify = 0;
  regmatch_t pmatch[1];
  char *Text, *ReqDocName, *DesDocName;
      
  // Initialize?
  if (a_node == NULL)
    {
      Options->VerifyParentWalkLevel = 1000;
      Options->VerifyChildWalkLevel = 1000;
      Options->FileTraceVerify = fopen (CsvFilename, "w");
      if (Options->FileTraceVerify == NULL)
        {
	  fprintf (stderr, "Error: Cannot create %s\n", CsvFilename);
	  return (1);
	}
      Options->TraceVerifyStateParent.NumRequirements = 0;
      Options->TraceVerifyState.NumRequirements = 0;
      fprintf (Options->FileTraceVerify, 
               "\"Requirement\",\"Requirement/Design Document Section\",\"Test Case\"\n");
      return (0);
    }
    
  if (Options->CombineSrdSdd)
    {
      ReqDocName = DesDocName = Options->ComDocName;
    }
  else
    {
      ReqDocName = Options->ReqDocName;
      if (!strcmp (Options->DesDocName, "DD"))
        DesDocName = "SDD";
      else
        DesDocName = Options->DesDocName;
    }
    
  // Process the node. ---------------------------------------------
  
  // Determine if it can contain a high-level requirement.  (Not 
  // necessarily that it DOES, just that it COULD.)  This will
  // be so if:
  //	1. It is part of the requirements document; and
  //	2. It matches the section-number constraint from the command
  //	   line switches.
  // By default, it is marked as NOT possible, so we only need to 
  // check the positive cases.
  if (Options->DocSections[REQ_DOC].Present)
    {
      if (Options->HighSectionRange == NULL)
        CanContainHighLevel = 1;	// No range constraint, so all okay!
      else
        {
	  // A section-number pattern has been defined by a command-line
	  // switch as a constraint on the sections that can contain
	  // requirements, so apply it and see if we're in the right
	  // range of section numbers.  Note that we make no effort
	  // to insure that the match encompasses the entire section
	  // number.  If the user desires that, then he should have
	  // prefixed and suffixed his pattern with '^' and '$' to 
	  // match the beginning of line and end of line.  If regexec
	  // returns 0, then it means a match was found.  
	  if (0 == regexec (&Options->HighSectionRangeMachine,
	  		    Options->DocSections[REQ_DOC].Number,
			    1, pmatch, 0))
	    CanContainHighLevel = 1;
	}
    }

  // If we've found a section that could contain requirements,
  // then we need to close out any existing requirements that 
  // we may have been accumulating test cases for.
  if (CanContainHighLevel)
    TraceVerifyCloseout (Options);

  // If could contain high-level requirements, let's check to see if it
  // DOES contain any.  If so, we have to buffer them.  We can't do anything
  // with them immediately, since we can only write them out when we have
  // some low-level requirements to go with them.
  if (CanContainHighLevel)
    {
#define TraceVerifyState punkin    
      Options->VerifyParentWalkLevel = Options->WalkLevel;
      Options->TraceVerifyStateParent.RequirementSection = Options->DocSections[REQ_DOC];
      strcpy (Options->TraceVerifyStateParent.DocName, ReqDocName);
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      printf("HL candidate %s, section %s:", Options->ReqDocName, 
             Options->DocSections[REQ_DOC].Number);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  We only check the so-called "section" language
      // matches if there are no "label" matches.
      for (i = j = 0; !regexec (&Options->HighLabelPatternMachine, &Text[j], 1, pmatch, 0); i++, j = pmatch[0].rm_eo)
        {
	  pmatch[0].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[0].rm_eo += j;
	  n = pmatch[0].rm_eo - pmatch[0].rm_so;
	  if (n >= MAX_REQUIREMENT_SIZE)
	    {
	      fprintf (stderr, "Error: Pattern produces string length >%d\n", MAX_REQUIREMENT_SIZE - 1);
	      xmlFree (Text);
	      return (1);
	    }
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  c = Text[pmatch[0].rm_eo];
	  Text[pmatch[0].rm_eo] = 0;
	  strcpy (Options->TraceVerifyStateParent.Requirements[i], &Text[pmatch[0].rm_so]);
	  Text[pmatch[0].rm_eo] = c;
	}
      for (; !i && !regexec (&Options->HighSectionPatternMachine, Text, 1, pmatch, 0); i++)
        {
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  strcpy (Options->TraceVerifyStateParent.Requirements[i], 
	          Options->DocSections[REQ_DOC].Number);
	}
      Options->TraceVerifyStateParent.NumRequirements = i;
      if (i == 0)
        {
          printf (" (none)");
	}
      else
	for (i = 0; i < Options->TraceVerifyStateParent.NumRequirements; i++)
	  printf (" %s", Options->TraceVerifyStateParent.Requirements[i]);
      printf ("\n");
      xmlFree (Text);   
      Options->VerifyChildWalkLevel = 1000;   
#undef TraceVerifyState      
    }

  ///////////////////////////////////////////////////////////////////////////////
  // Similarly, for the low-level requirements.  Note that unlike TraceDesign.c,
  // the low-level requirements are treated identically to the high-level 
  // level requirements, and actually overwrite the buffered data for them.
  if (Options->DocSections[DES_DOC].Present)
    {
      if (Options->LowSectionRange == NULL)
        CanContainLowLevel = 1;	// No range constraint, so all okay!
      else
        {
	  if (0 == regexec (&Options->LowSectionRangeMachine,
	  		    Options->DocSections[DES_DOC].Number,
			    1, pmatch, 0))
	    CanContainLowLevel = 1;
	}
    }
    
  // If we've found a section that could contain low-level requirements,
  // then we need to close out any existing requirements that 
  // we may have been accumulating test cases for.
  if (CanContainLowLevel)
    TraceVerifyCloseout (Options);

  if (CanContainLowLevel)
    {
      Options->VerifyChildWalkLevel = Options->WalkLevel;
      Options->TraceVerifyState.RequirementSection = Options->DocSections[DES_DOC];
      strcpy (Options->TraceVerifyState.DocName, DesDocName);
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      printf("LL candidate %s, section %s:", Options->DesDocName, 
             Options->DocSections[DES_DOC].Number, Text);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  We only check the so-called "section" language
      // matches if there are no "label" matches.
      for (i = j = 0; !regexec (&Options->LowLabelPatternMachine, &Text[j], 1, pmatch, 0); i++, j = pmatch[0].rm_eo)
        {
	  pmatch[0].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[0].rm_eo += j;
	  n = pmatch[0].rm_eo - pmatch[0].rm_so;
	  if (n >= MAX_REQUIREMENT_SIZE)
	    {
	      fprintf (stderr, "Error: Pattern produces string length >%d\n", MAX_REQUIREMENT_SIZE - 1);
	      xmlFree (Text);
	      return (1);
	    }
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  c = Text[pmatch[0].rm_eo];
	  Text[pmatch[0].rm_eo] = 0;
	  strcpy (Options->TraceVerifyState.Requirements[i], &Text[pmatch[0].rm_so]);
	  Text[pmatch[0].rm_eo] = c;
	}
      for (; !i && !regexec (&Options->LowSectionPatternMachine, Text, 1, pmatch, 0); i++)
        {
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  strcpy (Options->TraceVerifyState.Requirements[i], 
	          Options->DocSections[DES_DOC].Number);
	}
      Options->TraceVerifyState.NumRequirements = i;
      if (i == 0)
        {
          printf (" (none)");
	}
      else
	for (i = 0; i < Options->TraceVerifyState.NumRequirements; i++)
	  printf (" %s", Options->TraceVerifyState.Requirements[i]);
      printf ("\n");
      xmlFree (Text);      
    }

  ///////////////////////////////////////////////////////////////////////////////
  // Okay, now we need merely to act if this is an eligible section from the SVCP.
  if (Options->DocSections[VER_DOC].Present)
    {
      if (Options->VerifySectionRange == NULL)
        {
          fprintf (stderr, "Error: --verify-section-range needs to be defined.\n");
	  return (1);
	}
      else
        {
	  if (0 == regexec (&Options->VerifySectionRangeMachine,
	  		    Options->DocSections[VER_DOC].Number,
			    1, pmatch, 0))
	    CanContainVerify = 1;
	}
    }
    
  if (CanContainVerify)
    {
      TraceVerifyState_t *ts;
      
      if (Options->WalkLevel < Options->VerifyChildWalkLevel)
	{
	  // The test case must be associated with the high-level requirement.
	  ts = &Options->TraceVerifyStateParent;
	} 
      else
	{
	  // The test case must be associated with the low-level requirement.
	  ts = &Options->TraceVerifyState;
	}
	
      for (k = 0; k < ts->NumRequirements; k++)
	{
	  fprintf (Options->FileTraceVerify, "\"%s %s\",\"%s %s\",\"%s section %s\"\n",
		   ts->DocName,
		   ts->Requirements[k],
		   ts->DocName,
		   ts->RequirementSection.Number,
		   Options->VerDocName,
		   Options->DocSections[VER_DOC].Number);
	 }
      if (k == 0)	// Test case without requirements.
        {
	  fprintf (Options->FileTraceVerify, "\"(none)\",\"%s %s\",\"%s section %s\"\n",
		   ts->DocName,
		   ts->RequirementSection.Number,
		   Options->VerDocName,
		   Options->DocSections[VER_DOC].Number);
	}
      ts->TestCaseCount += ts->NumRequirements;
    }
  

  return (0);
}



