/*
  Name:        	WalkNodes.c
  Author:      	Ron Burkey
  Purpose:     	A function that walks through the nodes of an XML
  		structure created by libxml2 from a Do178Builder
		XML database.  The significant thing it does
		is to determine what DO-178B/DO-254 documents the
		nodes belong to, and the document section numbers
		associated with them.
  Mod History:	03/13/2008 RSB	Began.
  		03/19/2008 RSB	I was parsing the DO-254 input XML files
				wrong, in that I had the attribute names
				wrong.  (They're the same as for the 
				DO-178B files, namely "SRD", "DD", and so
				on, but I was using "HRD", "HDRD", etc.
				Inexplicably, it appeared to be working.
				Yikes!)
		03/21/2008 RSB	Added section titles.
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/

#include "Do178Builder-utility.h"
#include <ctype.h>

// This function prettifies a section title by undoing the special-character
// encoding that may have been done in it.
void
PrettifyTitle (const char *Raw, char *Pretty)
{
  // Fix leading digit of "_N_" to just "N".
  if (Raw[0] == '_' && isdigit (Raw[1]) && Raw[2] == '_')
    {
      *Pretty++ = Raw[1];
      Raw += 3;
    }
  // Step through now and find special characters.
  while (*Raw)
    {
      // "_rNr_" ?
      if (Raw[0] == '_' && Raw[1] == 'r' && isdigit (Raw[2]))
        {
	  const char *s;
	  unsigned u;
	  for (s = &Raw[3]; isdigit (*s); s++);
	  if (s[0] == 'r' && s[1] == '_')
	    {
	      sscanf (&Raw[2], "%u", &u);
	      if (u > ' ' && u < 127)
	        *Pretty++ = u;
	      else 
	        *Pretty++ = '.';
	      Raw = &s[2];
	      continue;
	    }
	}
      // "_"?
      if (*Raw == '_')
        {
          *Pretty++ = ' ';
	  Raw++;
	}
      else
        *Pretty++ = *Raw++;
    }
  // All done!
  *Pretty = 0;
}

// This function compares two SectionNumber_t structures, based on the numerical
// representation contained within.  (The string representation of the section
// number, if any, is ignored.)  Returns:
//	<0	if Section1 appears earlier in the document than Section2
//	0	if Section1 and Section2 are the same.
//	>0	if Section1 appears later in the document than Section2.
// This function isn't actually used within this file, but since this is the
// place that "general purpose" functions are defined, it seems a reasonable
// place to put it.
int
CompareSectionNumbers (SectionNumber_t *Section1, SectionNumber_t *Section2)
{
  int i, j;
  for (i = 0; i < Section1->Depth && i < Section2->Depth; i++)
    {
      j = Section1->Levels[i] - Section2->Levels[i];
      if (j != 0)
        return (j);
    }
  j = Section1->Depth - Section2->Depth;
  return (j);
}

// A function that can create a string representation from the numerical representation.
// Both the input and output are within a SectionNumber_t.
void
StringifySectionNumber (SectionNumber_t *Section, ProcessingOptions_t *Options)
{
  int i, j;
  char DecimalFormat[33] = "%d", DotDecimalFormat[33] = ".%d";

  if (Options->Sortable)
    {
      sprintf (DecimalFormat, "%c0%dd", '%', Options->Sortable);
      sprintf (DotDecimalFormat, ".%c0%dd", '%', Options->Sortable);
    }

  Section->Number[0] = 0;
  for (i = 0; i < Section->Depth; i++)
    {
      if (i == 0)
	j = sprintf (Section->Number, DecimalFormat, Section->Levels[i]);
      else
	j += sprintf (&Section->Number[j], DotDecimalFormat, Section->Levels[i]);
    }
}

// Similar to stringify, but with possible space padding of numerical
// fields to make them all come out even (and sortable!).
char * 
Outputtable (SectionNumber_t *Section, ProcessingOptions_t *Options)
{
  int i, jo;
  static char String[MAX_REQUIREMENT_SIZE];
  char DecimalFormat[33] = "%d", DotDecimalFormat[33] = ".%d";
  
  if (Options->Sortable)
    {
      sprintf (DecimalFormat, "%c%dd", '%', Options->Sortable);
      sprintf (DotDecimalFormat, ".%c%dd", '%', Options->Sortable);
    }
  String[0] = 0;
  for (i = 0; i < Section->Depth; i++)
    {
      if (i == 0)
	jo = sprintf (String, DecimalFormat, Section->Levels[i]);
      else
	jo += sprintf (&String[jo], DotDecimalFormat, Section->Levels[i]);
    }
  return (String);
}

// The following 3 functions are used to adjust section numbers when moving down to 
// children of the node (DocumentNodeDown), after returning from processing the 
// children (DocumentNodeUp), and when moving to a sibling node (DocumentNodeThis).

// Return 0 on success, non-zero on failure (section depth too great).
static int
DocumentNodeDown (xmlNode *a_node, ProcessingOptions_t *Options, char *DocName, int DocNum, char *DocName2)
{
  int i, j;
  SectionNumber_t *Section;
  
  Section = &Options->DocSections[DocNum];
  
  if (xmlHasProp (a_node, (xmlChar *) DocName) || (DocName2 != NULL && xmlHasProp (a_node, (xmlChar *) DocName2)))
    {
      if (Section->Depth >= MAX_SECTION_DEPTH)
	return (1);
      Section->Levels[Section->Depth] = 0;
      Section->Depth++;
    }
  return (0);
}	  
    
static void
DocumentNodeUp (xmlNode *a_node, ProcessingOptions_t *Options, char *DocName, int DocNum, char *DocName2)
{
  int i, j;
  SectionNumber_t *Section;
  
  Section = &Options->DocSections[DocNum];
  
  if (xmlHasProp (a_node, (xmlChar *) DocName) || (DocName2 != NULL && xmlHasProp (a_node, (xmlChar *) DocName2)))
    {
      if (Section->Depth > 1)
        Section->Depth--;
    }
}	  
    
static void
DocumentNodeThis (xmlNode *a_node, ProcessingOptions_t *Options, char *DocName, int DocNum, char *DocName2)
{
  SectionNumber_t *Section;
  
  Section = &Options->DocSections[DocNum];  
  Section->Title = (char *) a_node->name;
  
  if (Section->Depth == 0)
    {
      Section->Depth = 1;
      Section->Levels[0] = 0;
    }
  
  //printf ("%s\n", DocName);
  if (xmlHasProp (a_node, (xmlChar *) DocName) || (DocName2 != NULL && xmlHasProp (a_node,(xmlChar *)  DocName2)))
    {
      Section->Present = 1;
      Section->Levels[Section->Depth - 1]++;
      // Convert the section number from the Levels array to string form.
      StringifySectionNumber (Section, Options);
    }
  else
    Section->Present = 0;
}	  
    
// Walk through the nodes of an XML tree stored in memory, and 
// execute a function at each node.  Returns 0 on success, non-zero
// on error.  On detecting an error (with the function f), immediately
// aborts (and Options->WalkLevel will contain the depth at which the
// error occurred).  This function takes care of common activities 
// such as determining the documents and document section numbers
// associated with the node.
int
WalkNodes (xmlNode * a_node, NodeFunction_t *f, ProcessingOptions_t *Options)
{
  static int WalkLevel = 0;
  xmlNode *cur_node = NULL;
  
  WalkLevel++;
  Options->WalkLevel = WalkLevel;

  for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
    {
      if (cur_node->type == XML_ELEMENT_NODE)  
        {
	  // Figure out what docs this relates to and adjust section number.
	  // Note that the attribute tags for the requirements, design, and
	  // test-case documents are *always* "SRD", "SDD", and "SVCP", even
	  // if --DO-254 is used.  I was wrong about this at first!
	  if (Options->CombineSrdSdd)
	    {
	      DocumentNodeThis (cur_node, Options, "SRD", REQ_DOC, "DD");
	      Options->DocSections[DES_DOC] = Options->DocSections[REQ_DOC];
	      DocumentNodeThis (cur_node, Options, "SVCP", VER_DOC, NULL);
	    }
	  else
	    {
	      DocumentNodeThis (cur_node, Options, "SRD", REQ_DOC, NULL);
	      DocumentNodeThis (cur_node, Options, "DD", DES_DOC, NULL);
	      DocumentNodeThis (cur_node, Options, "SVCP", VER_DOC, NULL);
	    }
	    
          // Do specific processing on the node.
	  if ((*f) (cur_node, Options))
	    goto Error;	    
	}

      // Figure out what docs this relates to and adjust section number.
      if (Options->CombineSrdSdd)
	{
	  if (DocumentNodeDown (cur_node, Options, "SRD", REQ_DOC, "DD"))
	    goto Error;
	  if (DocumentNodeDown (cur_node, Options, "SVCP", VER_DOC, NULL))
	    goto Error;
	}
      else
	{
	  if (DocumentNodeDown (cur_node, Options, "SRD", REQ_DOC, NULL))
	    goto Error;
	  if (DocumentNodeDown (cur_node, Options, "DD", DES_DOC, NULL))
	    goto Error;
	  if (DocumentNodeDown (cur_node, Options, "SVCP", VER_DOC, NULL))
	    goto Error;
	}
      if (WalkNodes (cur_node->children, f, Options))
        return (1);
	
      // Figure out what docs this relates to and adjust section number.
      if (Options->CombineSrdSdd)
	{
	  DocumentNodeUp (cur_node, Options, "SRD", REQ_DOC, "DD");
	  DocumentNodeUp (cur_node, Options, "SVCP", VER_DOC, NULL);
	}
      else
	{
	  DocumentNodeUp (cur_node, Options, "SRD", REQ_DOC, NULL);
	  DocumentNodeUp (cur_node, Options, "DD", DES_DOC, NULL);
	  DocumentNodeUp (cur_node, Options, "SVCP", VER_DOC, NULL);
	}
    }
    
  WalkLevel--;
  Options->WalkLevel = WalkLevel;
  
  return (0);
Error:
  WalkLevel = 0;
  return (1);
}

